use strict;
use warnings;
package AutoDiff;

use Carp;

sub new{                           # the constructor
  my $class = $_[0];
  my $var;
  if (@_ == 3) {
    $var = [ $_[1], $_[2] ];           # 2 arguments [u, u']
  }
  elsif (@_ == 2) {
    $var = [ $_[1], 1 ];               # 1 argument [u, 1]: var
  }
  else {
    $var = [ 0, 0 ];                   # no argument [0,0]: var
  }
  bless $var, $class;
}

sub get_val {                      # accessor to the value
  $_[0]->[0];
}

sub set_val {
  $_[0]->[0] = $_[1];
}

sub get_der {                      # accessor to the derivative
  $_[0]->[1];
}

sub set_der {
  $_[0]->[1] = $_[1];
}


sub as_string{
  my $self = shift;
  return "($self->[0] , $self->[1])";
}

sub _add {                                  # addition
  my ($u, $v) = @_;
  $v = AutoDiff->new($v, 0) unless ref $v;
  AutoDiff->new(
    $u->[0] + $v->[0],
    $u->[1] + $v->[1]
  );
}

sub _sub {                                  # substraction
  my ($u, $v, $mut) = @_;
  $v = AutoDiff->new($v, 0) unless ref $v;
  if ($mut) {
    $mut=$u; $u=$v; $v=$mut;                  # swap $u and $v ($mut temp. var)
  }
  AutoDiff->new(
    $u->[0] - $v->[0],
    $u->[1] - $v->[1]
  );
}

sub _mul {                                  # multiplication
  my ($u, $v) = @_;
  $v = AutoDiff->new($v, 0) unless ref $v;
  AutoDiff->new(
    $u->[0] * $v->[0],
    $u->[1] * $v->[0] + $u->[0] * $v->[1]
  );
}

sub _div {                                  # division
  my ($u, $v, $mut) = @_;
  $v = AutoDiff->new($v, 0) unless ref $v;
  if ($mut) {
    $mut=$u; $u=$v; $v=$mut;                       # swap $u and $v ($mut temp. var)
  }
  AutoDiff->new(
    $u->[0] / $v->[0],
    ($u->[1] - $u->[0]*$v->[1]/$v->[0])/$v->[0]
  );
}

sub _pow {                                  # power **
  my ($u, $v, $mut) = @_;
  my $r = AutoDiff->new();
  $v = AutoDiff->new($v, 0) unless ref $v;
  if ($mut) {
    $mut=$u; $u=$v; $v=$mut;                  # swap $u and $v ($mut temp. var)
  }
  $r->[0] =  $u->[0]**$v->[0];
  if ($u->[0] > 0) {
    $r->[1] =  ($v->[1]*log($u->[0])+$v->[0]/$u->[0]*$u->[1])*$r->[0];
  }
  else {
    croak "Illegal power exponent" if $v->[1]!=0 or $v->[0]!=int($v->[0]);
    $r->[1] = $v->[0]*$u->[0]**($v->[0]-1)*$u->[1];
  }
  $r;
}
sub _sin {
  my $x = shift;
  AutoDiff->new(
    sin($x->[0]),
    cos($x->[0])*$x->[1]
  );
}

sub _cos {
  my $x = shift;
  AutoDiff->new(
    cos($x->[0]),
    -sin($x->[0])*$x->[1]
  );
}

sub _sqrt {
  my $x = shift;
  AutoDiff->new(
    sqrt($x->[0]),
    $x->[1]/2/sqrt($x->[0])
  );
}

sub _exp {
  my $x = shift;
  AutoDiff->new(
    exp($x->[0]),
    exp($x->[0])*$x->[1]
  );
}

sub _log {
  my $x = shift;
  AutoDiff->new(
    log($x->[0]),
    $x->[1]/$x->[0]
  );
}

use overload
  '""'   => 'as_string',
  '+'    => '_add',
  '-'    => '_sub',
  '*'    => '_mul',
  '/'    => '_div',
  '**'   => '_pow',
  'sin'  => '_sin',
  'cos'  => '_cos',
  'sqrt' => '_sqrt',
  'exp'  => '_exp',
  'log'  => '_log';
1;
